use std::{
    collections::hash_map::DefaultHasher,
    fmt::{Display, Write},
    hash::{Hash, Hasher},
};

use rand::{rngs::StdRng, thread_rng, Rng, SeedableRng};

use crate::{CakeSize, Centralness, Direction, Game, GameArgs, Icing, Pours, SimulationPen};

impl CakeSize {
    fn get_limits(&self) -> (u64, u64) {
        use CakeSize::*;
        match self {
            Small => (1, 6),
            Medium => (36, 10),
            Large => (100, 20),
        }
    }
}

impl Icing {
    fn sample_budget(&self, rng: &mut impl Rng, c: u64) -> u64 {
        use Icing::*;
        let target = match self {
            Little => c,
            Medium => 9 * c,
            Much => 243 * c,
        };
        sample_hyperbolic_pdf(rng, (target / 2).max(1), (target * 2).max(2))
    }
}

impl Pours {
    fn sample_max_pens(&self, rng: &mut impl Rng, c: u64) -> u64 {
        use Pours::*;
        let target;
        match self {
            Few => {
                target = c.min((1. + (c as f32).sqrt()).round() as u64);
            }
            Medium => {
                target = c.min((1. + (c as f32).sqrt() + c as f32 * 0.1).round() as u64);
            }
            Many => {
                target = c;
            }
        }
        sample_hyperbolic_pdf(rng, (target / 2).max(1), (target * 2).max(2))
    }
}

fn get_valuability(x: u64, y: u64, seed: u64, width: u64, height: u64, args: &GameArgs) -> i32 {
    let field_seed = seed ^ (y * width + x);
    let mut rng = StdRng::seed_from_u64(field_seed);
    // This is (hopefully) a hash and converting it to a float at the same time.
    let raw: f32 = rng.gen();
    use Centralness::*;
    let num_cells = width as f32 * height as f32;
    let a: f32 = match args.centralness {
        Spread => 1.,
        Medium => 1.0f32.max(num_cells / (0.3 * num_cells + 2.)),
        Concentrated => 1.0f32.max(num_cells / (0.1 * num_cells + 2.)),
    };
    let adjusted = 0.0f32.max(1. - a * (1. - raw));
    (adjusted * 1e3).round() as i32
}

#[derive(Debug, Clone, PartialEq)]
enum ClaimedStatus {
    Claimed { for_player_id: u8, strength: u32 },
    Unclaimed,
}

impl PartialOrd for ClaimedStatus {
    fn gt(&self, other: &Self) -> bool {
        if let Self::Claimed { strength, .. } = self {
            match other {
                Self::Unclaimed => true,
                Self::Claimed {
                    strength: other_strenght,
                    ..
                } => strength > other_strenght,
            }
        } else {
            false
        }
    }
    fn partial_cmp(&self, _other: &Self) -> Option<std::cmp::Ordering> {
        unimplemented!()
    }
}

fn sample_hyperbolic_pdf<R: Rng>(rng: &mut R, lb: u64, ub: u64) -> u64 {
    (rng.gen_range((lb as f64).ln()..=(ub as f64).ln())
        .exp()
        .round() as u64)
        .clamp(lb, ub - 1)
}

impl Game {
    pub fn from_map_args(mut args: GameArgs) -> Self {
        let mut seed = args.seed.unwrap_or_else(|| thread_rng().gen());
        if args.seed.is_none() {
            println!("Using seed {}", seed);
        }
        let mut hasher = DefaultHasher::new();
        seed.hash(&mut hasher);
        // Make sure passing the seed printed by "Using seed xyz" actually produces same result
        args.seed = None;
        args.hash(&mut hasher);
        seed = hasher.finish();

        let mut rng = StdRng::seed_from_u64(seed);

        let mut width = 0;
        let mut height = 0;
        let (minimum_area, maximum_side) = args.size.get_limits();
        while width * height < minimum_area {
            width = rng.gen_range(1..=maximum_side);
            height = rng.gen_range(1..=maximum_side);
        }

        // Change seed according to width and height because get_valuability is a bit weak
        width.hash(&mut hasher);
        height.hash(&mut hasher);
        seed = hasher.finish();

        let mut map = vec![];
        let mut c = 0;
        while c == 0 {
            map = (0..width)
                .map(|x| {
                    (0..height)
                        .map(|y| get_valuability(x, y, seed, width, height, &args))
                        .collect::<Vec<_>>()
                })
                .collect::<Vec<_>>();

            c = map
                .iter()
                .map(|col| {
                    col.iter()
                        .map(|cell| if *cell > 0 { 1 } else { 0 })
                        .sum::<u64>()
                })
                .sum();
            seed.hash(&mut hasher);
            seed = hasher.finish();
        }

        let budget = args.icing.sample_budget(&mut rng, c);
        let max_pens = args.pours.sample_max_pens(&mut rng, c);

        Self {
            map,
            budget,
            max_pens,
        }
    }
    pub fn simulate(&self, pens: &[SimulationPen], verbose: bool) -> (i64, i64) {
        let mut claim_statuses =
            vec![vec![ClaimedStatus::Unclaimed; self.map[0].len()]; self.map.len()];
        for sim_pen in pens {
            let mut x = sim_pen.pen.x;
            let mut y = sim_pen.pen.y;
            let mut dir_it = sim_pen.pen.moves.iter();
            let this_claim_status = ClaimedStatus::Claimed {
                for_player_id: sim_pen.owner_id,
                strength: sim_pen.pen.strength,
            };
            loop {
                if this_claim_status.gt(&claim_statuses[x][y]) {
                    claim_statuses[x][y] = this_claim_status.clone();
                }
                if let Some(dir) = dir_it.next() {
                    use Direction::*;
                    match *dir {
                        Up => y += 1,
                        Right => x += 1,
                        Down => y -= 1,
                        Left => x -= 1,
                    }
                } else {
                    break;
                }
            }
        }
        let mut res = (0, 0);
        if verbose {
            println!("End result:")
        }
        for y in (0..self.map[0].len()).rev() {
            for x in 0..self.map.len() {
                if let ClaimedStatus::Claimed { for_player_id, .. } = claim_statuses[x][y] {
                    match for_player_id {
                        1 => res.0 += self.map[x][y] as i64,
                        2 => res.1 += self.map[x][y] as i64,
                        _ => panic!("invalid player id"),
                    }
                    if verbose {
                        print!("{} ", for_player_id)
                    }
                } else if verbose {
                    print!("_ ");
                }
            }
            if verbose {
                println!();
            }
        }
        res
    }
}

impl Display for Game {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let (w, h) = (self.map.len(), self.map[0].len());
        f.write_fmt(format_args!(
            "{} {} {} {}\n",
            w, h, self.budget, self.max_pens
        ))?;
        for y in (0..h).rev() {
            for x in 0..w {
                f.write_fmt(format_args!("{} ", self.map[x][y]))?;
            }
            f.write_char('\n')?;
        }
        Ok(())
    }
}
