pub mod game;
pub mod pen;
pub mod player;

use std::{
    process::{Child, ChildStdin},
    sync::mpsc::Receiver,
    time::Instant,
};

use serde::{Deserialize, Serialize};

#[derive(Debug, thiserror::Error, Clone, Copy)]
pub enum BotError {
    #[error("Bot failed to properly print pour")]
    PenParseError,
    #[error("Bot attempted a pour which would require more icing than is left")]
    BudgetExceeded,
    #[error("Bot poured off the cake")]
    DisallowedMove,
    #[error("Bot experienced a runtime error")]
    ChildCrashed,
    #[error("Bot exited while game wasn't finished")]
    EarlyTermination,
    #[error("Bot used up time limit")]
    TimeLimitExceeded,
    // Not implemented
    #[error("Bot expected input while it should have printed something")]
    DeadLock,
}

#[derive(Serialize, Deserialize)]
pub enum Direction {
    Up,
    Right,
    Down,
    Left,
}

#[derive(Serialize, Deserialize)]
pub struct Pen {
    x: usize,
    y: usize,
    strength: u32,
    length: u64,
    moves: Vec<Direction>,
}

#[derive(Serialize, Deserialize)]
pub struct SimulationPen {
    pub pen: Pen,
    // 1 for player 1, 2 for player 2
    pub owner_id: u8,
    pub origin_turn: u64,
}

#[derive(Debug, Clone, Hash, clap::ValueEnum)]
enum Centralness {
    Spread,
    Medium,
    Concentrated,
}

#[derive(Debug, Clone, Hash, clap::ValueEnum)]
enum CakeSize {
    Small,
    Medium,
    Large,
}

#[derive(Debug, Clone, Hash, clap::ValueEnum)]
enum Icing {
    Little,
    Medium,
    Much,
}

#[derive(Debug, Clone, Hash, clap::ValueEnum)]
enum Pours {
    Few,
    Medium,
    Many,
}

#[derive(clap::Parser, Debug, Hash)]
pub struct GameArgs {
    #[clap(long)]
    centralness: Centralness,
    #[clap(long)]
    size: CakeSize,
    #[clap(long)]
    seed: Option<u64>,

    #[clap(long)]
    icing: Icing,
    #[clap(long)]
    pours: Pours,
}

#[derive(Serialize, Deserialize)]
pub struct Game {
    map: Vec<Vec<i32>>,
    budget: u64,
    pub max_pens: u64,
}

#[derive(Serialize, Deserialize)]
pub struct GameLog {
    pub game: Game,
    pub log_a: Vec<SimulationPen>,
    pub log_b: Vec<SimulationPen>,
    pub name_1: String,
    pub name_2: String,
}

#[derive(Debug)]
pub struct Player {
    budget: u64,

    child: Child,
    stdout_reciever: Receiver<Result<String, BotError>>,
    stdin: ChildStdin,
    runtime: f32,
    currently_waiting_since: Option<Instant>,
}
