use clap::Parser;
use fs_err as fs;

use judge::{BotError, Game, GameArgs, GameLog, Player, SimulationPen};

#[derive(clap::Parser, Debug)]
struct Args {
    #[clap(long)]
    command1: String,
    #[clap(long)]
    command2: String,
    #[clap(long)]
    name1: String,
    #[clap(long)]
    name2: String,
    #[clap(long)]
    visualization_output: Option<String>,
    #[clap(long, short)]
    verbose: bool,
    #[clap(flatten)]
    map_args: GameArgs,
}

fn unwrap_or_report_for_player<T>(result: Result<T, BotError>, player_id: u8) -> Option<T> {
    match result {
        Ok(t) => Some(t),
        Err(e) => {
            println!("Player {} caused failure: {}", player_id, e);
            None
        }
    }
}

// Takes two player commands and their corresponding id (1/2).
// Returns (score of played with id 1, score of player with id 2)
fn run_game(
    game: &Game,
    (player_a_command, player_a_id): (&str, u8),
    (player_b_command, player_b_id): (&str, u8),
    verbose: bool,
) -> Option<(Vec<SimulationPen>, (i64, i64))> {
    if verbose {
        println!(
            "Starting game of Player {} vs Player {}",
            player_a_id, player_b_id
        );
    }
    let mut player_a = unwrap_or_report_for_player(
        Player::from_command_str(player_a_command, game, true),
        player_a_id,
    )?;
    let mut player_b = unwrap_or_report_for_player(
        Player::from_command_str(player_b_command, game, false),
        player_b_id,
    )?;
    let mut pens = vec![];
    for turn in 0..2 * game.max_pens {
        let (mut on_move, mut other) = ((&mut player_a, player_a_id), (&mut player_b, player_b_id));
        if turn % 2 == 1 {
            std::mem::swap(&mut on_move, &mut other);
        }
        let pen = unwrap_or_report_for_player(on_move.0.read_pen(game), on_move.1)?;

        if verbose {
            println!("Player {} played pen: {}", on_move.1, pen);
        }

        unwrap_or_report_for_player(other.0.tell_pen(&pen), other.1)?;
        if !pen.is_empty() {
            pens.push(SimulationPen {
                pen,
                owner_id: on_move.1,
                origin_turn: turn,
            });
        }
    }
    let res = game.simulate(&pens, verbose);
    if verbose {
        println!(
            "Result of game: Player {}: {}, Player {}: {}",
            player_a_id, res.0, player_b_id, res.1
        );
    }
    Some((pens, res))
}

fn main() {
    let args = Args::parse();
    let game = Game::from_map_args(args.map_args);

    if args.verbose {
        println!("Game description: {}", game);
    }

    let (log_a, outcome_1) = match run_game(
        &game,
        (&args.command1, 1),
        (&args.command2, 2),
        args.verbose,
    ) {
        Some(res) => res,
        None => return,
    };
    let (log_b, outcome_2) = match run_game(
        &game,
        (&args.command2, 2),
        (&args.command1, 1),
        args.verbose,
    ) {
        Some(res) => res,
        None => return,
    };

    if let Some(visualization_output) = args.visualization_output {
        let game_log = GameLog {
            game,
            log_a,
            log_b,
            name_1: args.name1,
            name_2: args.name2,
        };
        let writer =
            fs::File::create(&visualization_output).expect("Failed to create log output file");
        serde_json::to_writer_pretty(writer, &game_log).expect("Failed to serialize game logs");
    }

    use std::cmp::Ordering::*;
    match (outcome_1.0 + outcome_2.0).cmp(&(outcome_1.1 + outcome_2.1)) {
        Less => println!("Winner is Player 2"),
        Equal => println!("Tie"),
        Greater => println!("Winner is Player 1"),
    }
}
