use std::{
    io::{Read, Write},
    process::{Child, Command, Stdio},
    str::FromStr,
    sync::mpsc,
    thread::{self, sleep},
    time::{Duration, Instant},
};

#[derive(Debug)]
struct StdoutIter<'a> {
    player: &'a mut Player,
    stdout: String,
    at: usize,
}

impl<'a> StdoutIter<'a> {
    fn new(player: &'a mut Player) -> Self {
        Self {
            player,
            stdout: String::new(),
            at: 0,
        }
    }
    fn append_stdout(&mut self) -> Result<(), BotError> {
        self.stdout.push_str(&self.player.wait_for_stdout()?);
        Ok(())
    }
    fn next_naive(&mut self) -> Option<&str> {
        let remaining = &self.stdout[self.at..];
        let after_whitespace = remaining.trim_start();
        let ret = after_whitespace.split_whitespace().next();
        self.at +=
            (remaining.len() - after_whitespace.len()) + ret.map(|ret| ret.len()).unwrap_or(0);
        ret
    }
    fn next(&mut self) -> Result<&str, BotError> {
        while self.stdout[self.at..].trim_start().is_empty() {
            self.append_stdout()?;
        }
        Ok(self.next_naive().unwrap())
    }
}

use crate::{BotError, Direction, Game, Pen, Player};

const TIME_LIMIT: f32 = 4.;

impl Drop for Player {
    fn drop(&mut self) {
        if let Err(e) = self.child.kill() {
            println!("Encountered error when killing child: {:?}", e);
        }
    }
}

fn check_early_termination(child: &mut Child) -> Result<(), BotError> {
    match child.try_wait().unwrap() {
        None => Ok(()),
        Some(status) => {
            if status.success() {
                Err(BotError::EarlyTermination)
            } else {
                Err(BotError::ChildCrashed)
            }
        }
    }
}

impl Player {
    pub fn from_command_str(
        command: &str,
        game: &Game,
        starting_player: bool,
    ) -> Result<Self, BotError> {
        let mut command = if let Some((program, args)) = command.split_once(' ') {
            let mut command = Command::new(program);
            command.args(args.split(' '));
            command
        } else {
            Command::new(command)
        };
        let mut child = command
            .stdin(Stdio::piped())
            .stdout(Stdio::piped())
            .spawn()
            .expect("Failed to start child process");
        println!("Started child with PID {}", child.id());
        let mut stdout = child.stdout.take().unwrap();
        let (sender, stdout_reciever) = mpsc::channel();
        thread::spawn(move || {
            let mut try_get_next_string = || -> Result<String, BotError> {
                let mut buf = vec![0; 64];
                let mut read_len = 0;
                while read_len == 0 {
                    read_len = stdout
                        .read(&mut buf)
                        // The child exiting early is handled later down the line.
                        .unwrap_or_default();
                    // It (sometimes?) just imagines 0s if there is nothing to be read
                    while read_len > 0 && buf[read_len - 1] == 0 {
                        read_len -= 1;
                    }
                    sleep(Duration::from_micros(100));
                }
                String::from_utf8(buf).map_err(|_| BotError::PenParseError)
            };
            loop {
                if sender.send(try_get_next_string()).is_err() {
                    break;
                };
            }
        });
        let mut stdin = child.stdin.take().unwrap();
        let _ = stdin.write_fmt(format_args!("{}", game));
        let _ = stdin.write_fmt(format_args!(
            "{}\n",
            if starting_player { "Y" } else { "O" }
        ));
        check_early_termination(&mut child)?;
        let mut currently_waiting_since = None;
        if starting_player {
            currently_waiting_since = Some(Instant::now());
        }
        Ok(Self {
            child,
            stdout_reciever,
            stdin,
            budget: game.budget,
            runtime: 0.,
            currently_waiting_since,
        })
    }
    pub fn tell_pen(&mut self, pen: &Pen) -> Result<(), BotError> {
        // This operation is allowed to fail, the bot doesn't have to live anymore if it did all
        // its duties.
        // TODO: only allow specific errors
        let _ = self.stdin.write_fmt(format_args!("{}", pen));
        match check_early_termination(&mut self.child) {
            // It can be ok for bots to exit early if they printed all their pens
            Err(BotError::EarlyTermination) => Ok(()),
            v => v,
        }?;
        self.currently_waiting_since = Some(Instant::now());
        Ok(())
    }
    fn cpu_time(&self) -> Result<f32, BotError> {
        Ok(self.runtime
            + self
                .currently_waiting_since
                .map(|start| start.elapsed().as_secs_f32())
                .unwrap_or_default())
    }

    fn wait_for_stdout(&mut self) -> Result<String, BotError> {
        let mut error_count = 0;
        loop {
            if self.cpu_time()? > TIME_LIMIT {
                return Err(BotError::TimeLimitExceeded);
            }
            if let Ok(ret) = self.stdout_reciever.try_recv() {
                return ret;
            }
            // It is important that this only happens if we would have to wait for stdout.
            match check_early_termination(&mut self.child) {
                Ok(()) => {}
                Err(e) => {
                    error_count += 1;
                    // This is a bit ugly but it is needed because if the bots prints the last pen
                    // and then immediately exits, it sometmies takes the stdout reader another
                    // moment to actually read the pen and if we didn't wait for that, we might
                    // falsely diagnose an early sucessful exit.
                    if error_count == 10 {
                        return Err(e);
                    }
                }
            }
            sleep(Duration::from_micros(100));
        }
    }

    pub fn read_pen(&mut self, game: &Game) -> Result<Pen, BotError> {
        let mut f = || {
            // self is unusable while the iter exists so we copy the budget before
            let budget = self.budget;
            let mut bot_stdout_iter = StdoutIter::new(self);
            fn try_parse_next<T: FromStr>(it: &mut StdoutIter) -> Result<T, BotError> {
                it.next()?.parse().map_err(|_| BotError::PenParseError)
            }

            let x: u64 = try_parse_next(&mut bot_stdout_iter)?;
            let y: u64 = try_parse_next(&mut bot_stdout_iter)?;
            let strength = try_parse_next(&mut bot_stdout_iter)?;
            let length: u64 = try_parse_next(&mut bot_stdout_iter)?;
            if strength == 0 {
                return Err(BotError::PenParseError);
            }
            let costf = length as f64 * 3.0f64.powi(strength as i32 - 1);
            if costf - 1000. > budget as f64 {
                return Err(BotError::BudgetExceeded);
            }
            let cost = length * 3u64.pow(strength - 1);
            if cost > budget {
                return Err(BotError::BudgetExceeded);
            }
            let moves: Vec<Direction> = (0..length.max(1) - 1)
                .map(|_| try_parse_next(&mut bot_stdout_iter))
                .collect::<Result<_, _>>()?;

            // Read everything that needs to be read from the bot
            self.runtime += self
                .currently_waiting_since
                .unwrap()
                .elapsed()
                .as_secs_f32();
            self.currently_waiting_since = None;

            let (w, h) = (game.map.len() as u64, game.map[0].len() as u64);
            let mut check_x = x;
            let mut check_y = y;
            let mut move_it = moves.iter();
            loop {
                if check_x.clamp(0, w - 1) != check_x || check_y.clamp(0, h - 1) != check_y {
                    return Err(BotError::DisallowedMove);
                }
                if let Some(dir) = move_it.next() {
                    use Direction::*;
                    match *dir {
                        Up => check_y += 1,
                        Right => check_x += 1,
                        Down => check_y -= 1,
                        Left => check_x -= 1,
                    };
                } else {
                    break;
                }
            }
            self.budget -= cost;
            Ok(Pen {
                x: x as usize,
                y: y as usize,
                strength,
                length,
                moves,
            })
        };
        match f() {
            Err(BotError::PenParseError) => match check_early_termination(&mut self.child) {
                // If the bot crashed, the error message may be attempted to be parsed as a pen.
                Err(BotError::ChildCrashed) => Err(BotError::ChildCrashed),
                _ => Err(BotError::PenParseError),
            },
            ret => ret,
        }
    }
}
