import {
    init,
    classModule,
    propsModule,
    styleModule,
    eventListenersModule,
    attributesModule,
    h
} from "snabbdom";

const patch = init([
    classModule,
    propsModule,
    styleModule,
    attributesModule,
    eventListenersModule
]);

// returns the SVG coordinates of the pointer for an event
function eventSVGPoint(event, svg = event.currentTarget.ownerSVGElement || event.currentTarget) {
    const point = svg.createSVGPoint()
    point.x = event.clientX
    point.y = event.clientY
    return point.matrixTransform(svg.getScreenCTM().inverse())
}

function repeat(n, el) {
    const arr = new Array(n)
    for (let i = 0; i < n; i++) arr[i] = window.structuredClone(el)
    return arr
}

let PLAYER_1_HUE = 39
let PLAYER_2_HUE = 214

let mapViewScale = 0.7
let mapViewX = (1 / mapViewScale / 2 - 0.5) * mapViewScale
let mapViewY = (1 / mapViewScale / 2 - 0.5) * mapViewScale

let dragPt = null
const mapTouches = new Map()
let lastTwoFingerParams

let width
let height
let budget
let max_pens
let valuabilities
let valuability_sum
let claims
let pens
let pens_consequences
let pen_index
let phase
let current_game
let is_initialized = false

let hasFileUrl = false
let urlFetchError = null

let name_1
let name_2
let score_1
let score_2
let budget_1
let budget_2
let turn

function getTwoFingerParams() {
    const [a, b] = mapTouches.values()
    const dx = a.x - b.x
    const dy = a.y - b.y
    return {
        x: (a.x + b.x) / 2,
        y: (a.y + b.y) / 2,
        d: Math.sqrt(dx * dx + dy * dy),
    }
}

function zoomMap(anchor, factor) {
    const newScale = Math.max(0.01, Math.min(1, mapViewScale * (factor || 0)))
    mapViewX = anchor.x + (mapViewX - anchor.x) * newScale / mapViewScale
    mapViewY = anchor.y + (mapViewY - anchor.y) * newScale / mapViewScale
    mapViewScale = newScale
}

function startDrag(event) {
    event.currentTarget.setPointerCapture(event.pointerId)
    dragPt = eventSVGPoint(event)
}

function dragDelta(event) {
    const cursor = eventSVGPoint(event)
    const delta = { x: cursor.x - dragPt.x, y: cursor.y - dragPt.y }
    dragPt = cursor
    return delta
}

function onWheelMap(event) {
    event.preventDefault()
    if (event.ctrlKey) {
        mapViewX -= event.deltaX
        mapViewY -= event.deltaY
    } else {
        zoomMap(eventSVGPoint(event), Math.exp(-event.deltaY / 300))
    }
    render()
}

function onPointerdownMap(event) {
    event.preventDefault()
    if (document.activeElement != null) document.activeElement.blur()

    const cursor = eventSVGPoint(event)

    if (mapTouches.size === 0 && (event.button === 1 || event.button == 0)) {
        startDrag(event)
    }

    // Two-finger touch
    if (event.pointerType === 'touch') {
        if (mapTouches.size === 1) {
            dragPt = null
        }
        mapTouches.set(event.pointerId, cursor)
        event.currentTarget.setPointerCapture(event.pointerId)
        if (mapTouches.size === 2) {
            lastTwoFingerParams = getTwoFingerParams()
            render()
            return
        }
    }
}

function onPointermoveMap(event) {
    if (mapTouches.has(event.pointerId)) {
        mapTouches.set(event.pointerId, eventSVGPoint(event))
        if (mapTouches.size === 2) {
            const params = getTwoFingerParams()

            mapViewX += params.x - lastTwoFingerParams.x
            mapViewY += params.y - lastTwoFingerParams.y
            zoomMap(params, params.d / lastTwoFingerParams.d)

            lastTwoFingerParams = params
            render()
        }
    }

    if (dragPt != null) {
        const delta = dragDelta(event)

        mapViewX += delta.x
        mapViewY += delta.y

        render()
    }

}

function onPointerupMap(event) {
    if (mapTouches.has(event.pointerId)) {
        mapTouches.delete(event.pointerId)
        event.currentTarget.releasePointerCapture(event.pointerId)
    }
    dragPt = null
}

function inSvg(children) {
    return h(
        "svg.main-svg", {
        attrs: { viewBox: "0 0 1 1" }, on: {
            wheel: onWheelMap,
            pointerdown: onPointerdownMap,
            pointermove: onPointermoveMap,
            pointerup: onPointerupMap,
            pointercancel: onPointerupMap,
        }
    },
        [
            h("g", { attrs: { transform: `translate(${mapViewX}, ${mapViewY}) scale(${mapViewScale})` } }, children)
        ]
    );
}

async function onFileInput(event) {
    if (event.target.files.length >= 1) {
        const file = event.target.files[0]
        parseLogs(await file.text())
    }
}

function logFileLoadMenu() {
    return h("input.file-upload-btn", {
        attrs: { type: "file", accept: 'application/json,.json', },
        on: { change: onFileInput }
    });
}

function resetGameState() {
    pen_index = 0
    phase = 0
    claims = repeat(width, repeat(height, { strength: 0, owner: 0 }))
    score_1 = 0
    score_2 = 0
    budget_1 = budget
    budget_2 = budget
    turn = pens[current_game][pen_index].origin_turn + 1;
}

async function parseLogs(text) {
    let data = JSON.parse(text)
    valuabilities = data.game.map
    valuability_sum = 0
    for (let x = 0; x < valuabilities.length; x++) {
        for (let y = 0; y < valuabilities[x].length; y++) {
            valuability_sum += Math.max(0, valuabilities[x][y])
        }
    }
    width = valuabilities.length
    height = valuabilities[0].length
    pens = { "Game A": data.log_a, "Game B": data.log_b }
    pens_consequences = { "Game A": [], "Game B": [] }
    budget = data.game.budget
    max_pens = data.game.max_pens
    current_game = "Game A"
    name_1 = data.name_1
    name_2 = data.name_2
    resetGameState()
    is_initialized = true
    render()
}

function onGameChange(event) {
    current_game = event.target.childNodes[0].data
    resetGameState()
    render()
}

function gameMenu() {
    let A_br = 255
    let B_br = 255
    if (current_game == "Game A") {
        A_br = 230
    } else {
        B_br = 230
    }
    return h("div.game-selection.button-group", {}, [
        h("button.group-select-button", { on: { click: onGameChange }, style: { backgroundColor: `rgb(${A_br} ${A_br} ${A_br})` } }, "Game A"),
        h("button.group-select-button", { on: { click: onGameChange }, style: { backgroundColor: `rgb(${B_br} ${B_br} ${B_br})` } }, "Game B")
    ])
}

function coveredCells(pen) {
    let x = pen.x
    let y = pen.y
    let cells = [[x, y]]
    for (let i = 0; i < pen.length - 1; i++) {
        let direction = pen.moves[i]
        if (direction == "Right") {
            x += 1
        } else if (direction == "Up") {
            y += 1
        } else if (direction == "Left") {
            x -= 1
        } else if (direction == "Down") {
            y -= 1
        } else {
            alert("Invalid direction")
        }
        cells.push([x, y])
    }
    return cells
}

function calcNextConsequences() {
    let consequences = []
    let pen = pens[current_game][pen_index]
    let owner = pen.owner_id
    pen = pen.pen

    let covered_cells = coveredCells(pen)
    for (let i = 0; i < covered_cells.length; i++) {
        let [claiming_x, claiming_y] = covered_cells[i]
        if (claims[claiming_x][claiming_y].strength < pen.strength) {
            consequences.push({
                x: claiming_x, y: claiming_y,
                original: claims[claiming_x][claiming_y],
                new: { strength: pen.strength, owner: owner }
            })
        }
    }
    pens_consequences[current_game].push(consequences)
}

function updateClaim(x, y, new_val) {
    function updateScore(player, change) {
        if (player == 1) {
            score_1 += change
        } if (player == 2) {
            score_2 += change
        }
    }
    updateScore(claims[x][y].owner, -valuabilities[x][y])
    updateScore(new_val.owner, valuabilities[x][y])
    claims[x][y] = new_val
}

function applyConsequences(i, forward) {

    let budget_change = - pens[current_game][i].pen.length * 3 ** (pens[current_game][i].pen.strength - 1);
    if (!forward) {
        budget_change *= -1
    }
    if (pens[current_game][i].owner_id == 1) {
        budget_1 += budget_change
    } else {
        budget_2 += budget_change
    }

    for (let ci = 0; ci < pens_consequences[current_game][i].length; ci++) {
        const consequence = pens_consequences[current_game][i][ci]
        let new_val = consequence.new
        if (!forward) {
            new_val = consequence.original
        }
        updateClaim(consequence.x, consequence.y, new_val, forward)
    }
}

function playbackForward() {
    if (pen_index == pens[current_game].length) {
        return
    }
    if (phase == 0) {
        phase = 1
    } else {
        phase = 0
        if (pens_consequences[current_game].length == pen_index) {
            calcNextConsequences()
        }

        applyConsequences(pen_index, true)
        pen_index += 1
    }
    if (pen_index < pens[current_game].length) {
        turn = pens[current_game][pen_index].origin_turn + 1;
    }

    render()
}

function playbackBackward() {
    if (phase == 1) {
        phase = 0
    } else {
        if (pen_index == 0) {
            return
        }
        phase = 1
        pen_index -= 1
        applyConsequences(pen_index, false)
    }
    turn = pens[current_game][pen_index].origin_turn + 1;

    render()
}

function playbackRoundForward() {
    // renders twice but that's ok
    playbackForward()
    playbackForward()
}

function playbackRoundBackward() {
    // renders twice but that's ok
    playbackBackward()
    playbackBackward()
}

function playbackStart() {
    resetGameState()

    render()
}

function playbackEnd() {
    while (pen_index < pens[current_game].length) {
        if (pens_consequences[current_game].length == pen_index) {
            calcNextConsequences()
        }

        applyConsequences(pen_index, true)
        pen_index += 1
    }
    phase = 0

    turn = max_pens * 2
    render()
}

function ScoresIndicator() {
    let children = []
    let color_a = `hsl(${PLAYER_1_HUE} 100% 50%)`
    let color_b = `hsl(${PLAYER_2_HUE} 100% 50%)`
    let text_offset = 0.1

    svg_width = window.innerWidth / (window.innerHeight * 0.25) * 1.5

    function bar(y, part_a, part_b, scores) {
        let bar_children = []
        if (scores) {
            bar_children.push(h("rect", {
                attrs: {
                    x: svg_width * 0.1, y: y,
                    width: svg_width * 0.8, height: 0.15,
                    fill: "lightgray",
                }
            }))
        }

        bar_children.push(h("rect", {
            attrs: {
                x: svg_width * 0.1, y: y,
                width: svg_width * 0.8 * part_a, height: 0.15,
                fill: color_a,
            }
        }))

        bar_children.push(h("rect", {
            attrs: {
                x: svg_width * (0.9 - 0.8 * part_b), y: y,
                width: svg_width * 0.8 * part_b, height: 0.15,
                fill: color_b,
            }
        }))

        if (scores) {
            bar_children.push(h("rect", {
                attrs: {
                    x: svg_width * 0.499, y: y,
                    width: svg_width * 0.002, height: 0.15,
                    fill: "black",
                }
            }))
        }
        return h("g", {}, bar_children)
    }

    let y_of = 0.55

    children.push(h("defs", {}, h("filter", { attrs: { x: 0, y: 0, width: 1, height: 1, id: "bg-filter" } },
        [h("feFlood", { attrs: { "flood-color": "white", result: "bg", "flood-opacity": 0.6 } }),
        h("feMerge", {}, [
            h("feMergeNode", { attrs: { in: "bg" } }),
            h("feMergeNode", { attrs: { in: "SourceGraphic" } })
        ])
        ]
    )))

    // Turn number
    children.push(h("text", { attrs: { x: svg_width * 0.5, y: 0.48, "font-size": 0.2, "text-anchor": "middle", color: "black", filter: "url(#bg-filter)", stroke: "black", "stroke-width": 0.015 } }, `Turn ${turn}/${2 * max_pens}`))

    // Names
    children.push(h("text", { attrs: { x: svg_width * 0.1, y: 0.72, "font-size": 0.2, fill: color_a } }, `${name_1}`))
    children.push(h("text", { attrs: { x: svg_width * 0.9, y: 0.72, "font-size": 0.2, fill: color_b, "text-anchor": "end" } }, `${name_2}`))

    // Score indicators
    children.push(h("text", { attrs: { x: svg_width * 0.5, y: 0.72, "font-size": 0.2, "text-anchor": "middle", color: "black", filter: "url(#bg-filter)" } }, "Score"))
    children.push(h("text", { attrs: { x: svg_width * (0.5 - text_offset), y: 0.72, "font-size": 0.2, fill: color_a } }, `${score_1}`))
    children.push(h("text", { attrs: { x: svg_width * (0.5 + text_offset), y: 0.72, "font-size": 0.2, fill: color_b, "text-anchor": "end" } }, `${score_2}`))
    children.push(bar(0.8, Math.max(0, score_1) / valuability_sum, Math.max(0, score_2) / valuability_sum, true))

    // Budget indicator
    children.push(h("text", { attrs: { x: svg_width * 0.5, y: 1.17, "font-size": 0.2, "text-anchor": "middle", color: "black", filter: "url(#bg-filter)" } }, "Budget"))
    children.push(h("text", { attrs: { x: svg_width * (0.5 - text_offset), y: 1.17, "font-size": 0.2, fill: color_a } }, `${budget_1}`))
    children.push(h("text", { attrs: { x: svg_width * (0.5 + text_offset), y: 1.17, "font-size": 0.2, fill: color_b, "text-anchor": "end" } }, `${budget_2}`))
    children.push(bar(1.25, budget_1 / budget / 2, budget_2 / budget / 2, false))

    return h("svg.scores-svg", { attrs: { viewBox: "0 0 1 1.5", preserveAspectRatio: "xMinYMin" } }, children)
}

function PlaybackButtons() {
    return h('div.playback-nav', [
        h('button.nav-button', { on: { click: playbackStart }, attrs: { title: 'Go to start (Home)' } }, h('svg', h('path', { attrs: { d: 'M18 18l-8.5-6L18 6v12zM8 6v12h-2V6h2z' } }))),
        h('button.nav-button', { on: { click: playbackRoundBackward }, attrs: { title: 'Previous round (Page Up)' } }, h('svg', h('path', { attrs: { d: 'M11 18V6l-8.5 6 8.5 6zm.5-6l8.5 6V6l-8.5 6z' } }))),
        h('button.nav-button', { on: { click: playbackBackward }, attrs: { title: 'Previous frame (Left Arrow)' } }, h('svg', h('path', { attrs: { d: 'M15.5 18V6l-8.5 6 8.5 6z' } }))),
        h('button.nav-button', { on: { click: playbackForward }, attrs: { title: 'Next frame (Right Arrow)' } }, h('svg', h('path', { attrs: { d: 'M8.5 18l8.5-6L8.5 6v12z' } }))),
        h('button.nav-button', { on: { click: playbackRoundForward }, attrs: { title: 'Next round (Page Down)' } }, h('svg', h('path', { attrs: { d: 'M4 18l8.5-6L4 6v12zm9-12v12l8.5-6L13 6z' } }))),
        h('button.nav-button', { on: { click: playbackEnd }, attrs: { title: 'Go to end (End)' } }, h('svg', h('path', { attrs: { d: 'M6 18l8.5-6L6 6v12zM16 6v12h2V6h-2z' } }))),
    ]);
}

let MAX_VALUABILITY = 1e3
let MAX_STRENGTH = 13
function cell(valuability, xi, yi, claim_strength, claim_owner) {
    let s = 1.0 / Math.max(width, height)

    let children = [
        // Background
        h("rect", { attrs: { x: xi * s, y: 1 - (yi + 1) * s, width: s, height: s, fill: `#f0f0f0` } }),
    ]

    if (claim_strength > 0) {

        // Claim background
        let claim_hue = PLAYER_1_HUE
        if (claim_owner === 2) {
            claim_hue = PLAYER_2_HUE
        }
        let claim_l = 100 - Math.max(claim_strength + 1, MAX_STRENGTH / 2) / MAX_STRENGTH * 100
        let claim_opacity = Math.min(1, (claim_strength + 1) / MAX_STRENGTH * 2)

        children.push(
            h("rect", {
                attrs: {
                    x: xi * s,
                    y: 1 - (yi + 1) * s,
                    width: s, height: s,
                    fill: `hsl(${claim_hue} 100% ${claim_l}%)`,
                    "fill-opacity": claim_opacity
                }
            })
        )

        // Strength text
        let strength_labels = []
        let STRENGTH_C = 4.2312
        let font_s = s / STRENGTH_C
        if (("" + claim_strength).length == 2) {
            STRENGTH_C = Math.PI
        }
        let strength_s = s / STRENGTH_C
        let clip_id = "clip-" + xi + "-" + yi

        children.push(
            h("defs", {}, [
                h("clipPath", { attrs: { id: clip_id } }, [
                    h("rect", { attrs: { x: xi * s, y: 1 - (yi + 1) * s, width: s, height: s } })
                ])
            ])
        )

        for (let x = Math.floor((xi * s - strength_s) / strength_s) * strength_s; x - strength_s < (xi + 1) * s; x += strength_s) {
            for (let y = Math.floor((1 - (yi + 1) * s - strength_s) / strength_s) * strength_s; y - strength_s < 1 - yi * s; y += strength_s) {
                strength_labels.push(
                    h("text", {
                        attrs: {
                            x: x, y: y,
                            fill: `hsl(${claim_hue} 100% ${claim_l}%)`,
                            "font-size": font_s,
                            "text-anchor": "middle",
                            "dominant-baseline": "central",
                            stroke: "white", "stroke-width": font_s / 9,
                        }
                    }, [`${claim_strength}`])
                )
            }
        }

        children.push(h("g", { attrs: { "clip-path": `url(#${clip_id})` } }, strength_labels))
    }

    // Border
    children.push(
        h("rect", { attrs: { x: xi * s, y: 1 - (yi + 1) * s, width: s, height: s, fill: `none`, stroke: "black", "stroke-width": s / 20 } }),
    )

    if (valuability != 0) {
        // Valuability
        let val_s = Math.abs(valuability) / MAX_VALUABILITY * s * 0.6
        let val_fill = 'black'
        if (valuability < 0) {
            val_fill = 'red'
        }
        children.push(
            h("rect", {
                attrs: {
                    x: (xi + 0.5) * s - val_s / 2,
                    y: 1 - (yi + 0.5) * s - val_s / 2,
                    width: val_s, height: val_s,
                    fill: val_fill
                }
            })
        )
    }

    return h("g", {}, children);
}

function playingGrid() {
    let cells = []
    for (let x = 0; x < width; x++) {
        for (let y = 0; y < height; y++) {
            cells.push(cell(valuabilities[x][y], x, y, claims[x][y].strength, claims[x][y].owner))
        }
    }
    return cells
}

function penPreview() {
    if (phase == 1) {
        let pen = pens[current_game][pen_index]
        let children = []
        let cells = coveredCells(pen.pen)
        let color = `hsl(${PLAYER_1_HUE} 100% 50%)`
        if (pen.owner_id == 2) {
            color = `hsl(${PLAYER_2_HUE} 100% 50%)`
        }
        let s = 1.0 / Math.max(width, height)
        let t = 0.4 * s

        // In case pen.length = 0
        children.push(h("rect", {
            attrs: {
                x: (cells[0][0] + 0.5) * s - t / 2,
                y: 1 - (cells[0][1] + 0.5) * s - t / 2,
                height: t, width: t,
                fill: color
            }
        }))

        for (let i = 0; i < cells.length - 1; i++) {
            let from = cells[i]
            let to = cells[i + 1]
            let x = (from[0] + 0.5) * s - t / 2
            let y = 1 - (from[1] + 0.5) * s - t / 2
            let width = s + t
            let height = t
            if (to[1] != from[1]) {
                height = width
                width = t
            }
            if (to[0] < from[0]) {
                x -= s
            }
            if (to[1] > from[1]) {
                y -= s
            }
            children.push(h("rect", {
                attrs: {
                    x: x,
                    y: y,
                    height: height, width: width,
                    fill: color
                }
            }))
        }

        return h("g", { attrs: { opacity: 0.5 } }, children)

    } else {
        return h("g")
    }
}

function fetchError(error) {
    urlFetchError = error.message
    render()
}

const URL_PREFIX = '#url='
if (location.hash.startsWith(URL_PREFIX)) {
    hasFileUrl = true
    const fileUrl = location.hash.slice(URL_PREFIX.length)
    fetch(fileUrl).then(response => {
        if (!response.ok) {
            fetchError(new Error(`HTTP ${response.status}`))
        } else {
            response.text().then(
                text => parseLogs(text),
                error => fetchError(error))
        }
    }, error => fetchError(error))
}

let vnode = document.getElementById('container')

function render() {
    if (!is_initialized) {
        if (hasFileUrl) {
            // Todo: show errors
        } else {
            vnode = patch(vnode, h("div.log-input", {}, [h("p.padme-right", {}, ["Visualization log:"]), logFileLoadMenu()]))
        }
    } else {
        const newVnode = h('div.h-full', [ScoresIndicator(), gameMenu(), inSvg(playingGrid().concat(penPreview())), PlaybackButtons()])
        vnode = patch(vnode, newVnode)
    }

    let title = 'colorclaim visualization'
    if (document.title !== title) document.title = title
}

window.addEventListener('resize', render)

render()